﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Threading; // threading
using System.Runtime.InteropServices; // for GetLastError()

using AIOWDMNet;  // the namespace exposes the AIOWDM Class interface 


namespace COS
{

    public partial class Form1 : Form
    {
        Thread myThread;

        UInt32 Status = 0;

        // This is the basic data that we will be using to interface with the card:
        public UInt32 Offset = 0;   // This the offset that will be used based on the base address of the card
        public bool RunFlag = false;
        //public bool TimerEnabled = false;
        

        // Data used to discover cards in the system:
        const UInt32 MAX_CARDS = 10;
        static int NumCards = 0;
        static uint CardNum = 0;

        static int numIRQ = 0; //temp IRQ counter
        static int LastError = 0;
        static bool bAbort = false;

        // Data:
        public struct TCOSData
        {
            public byte A, B, C;
        }
        
        public struct TCardData 
        {
            public TCardData(int i)
            {
                mi = i;
                Base = 0;
                PPIs = 0;
                IsWDG = false;
                IRQCount = 0;
                Log = "NULL";
                WDGPets = 0;
            }
            public int mi;
            public ulong  Base;
            public ulong  PPIs;
            public bool   IsWDG;
            public ulong  IRQCount;
            public String     Log;
            public ulong  WDGPets;
	    }

        public TCardData CardData; // only one struct

        // Using MS Mutex example to manage threads:       
        // Create a new Mutex. The creating thread does not own the Mutex.
        private static Mutex mut = new Mutex();
        private const int numIterations = 1;
        private const int numThreads = 1;

        public Form1()
        {
            InitializeComponent();

        }
        
        private void Form1_Load(object sender, EventArgs e)
        {
    
            if (AIOWDM.InPortB(0x61) == 0xAA55) 
            {
                MessageBox.Show("ACCESNT.SYS not detected. Please copy ACCESNT.SYS into [NT]/system32/drivers and re-run this sample.", "Warning");
            }

            FindCardsWDM();  // this will get all the card info


            bool found = false;
            
            // Local Vars for QueryCardInfo():
            UInt32 DeviceID, Base;
            UInt32 NameSize = 256;
            UInt16[] Name = new UInt16[256];
            String strname;
            
            // Setting CardData to 0 as needed            
		    CardData.Base = 0;
		    CardData.PPIs = 0;
		    CardData.IRQCount = 0;
		    CardData.WDGPets = 0;
		    CardData.IsWDG = false;

            // Get the total number of cards installed:
            NumCards = AIOWDM.GetNumCards();
             
	        // Beginning Init for card 
            //char[] BaseHex = new char[10];
      
            if (NumCards == 0) 
	        {
                //labelCardName.Text = "No Card Found";
                strname = String.Format("No Card Found");
                comboBoxCard.Items.Add(strname);
                comboBoxCard.SelectedIndex = 0;

                textBoxMessage.Text = "No cards were found!";
                textBoxMessage.Text += "This may mean the card is not installed. ";
                textBoxMessage.Text += "Check Device Manager for a card and its status";
                textBoxMessage.Text += "You may consider rebooting your system.";
                comboBoxCard.Enabled = false;

                //RunFlag = false;
	        } 
	        else 
	        {
                for (CardNum = 0; CardNum <= NumCards-1; CardNum++)
                {
		          CardData.IsWDG = false;

                  // Validate and store card data:
                  Status = AIOWDM.QueryCardInfo(CardNum, out DeviceID, out Base, out NameSize, out strname);
		          		          
                  switch (DeviceID) 
		          {
			        case 0x0E50: //PCI-DIO-24S
			            CardData.PPIs = 1;
			            break;
			        case 0x0E56: //PCI-DIO-24DCS
			            CardData.PPIs = 1;
			            break;
			        case 0x0E60: //PCI-DIO-48S
			            CardData.PPIs = 2;

                        strname = String.Format("PCI-DIO-48S Parallel Digital I/O Card");
                        comboBoxCard.Items.Add(strname);
                        comboBoxCard.SelectedIndex = 0;
                        
                        CardData.Base = Base;
                        found = true;
			            break;
			        case 0x22C0: case 0x25C0: case 0x2FC0: case 0x2FC1: //PCI-WDG-CSM, P104-WDG-E, P104-WDG-CSM, and P104-WDG-CSMA
                        
                        strname = String.Format("PCI-WDG-CSM or P104 WDG Board");  // Port 0 ??
                        comboBoxCard.Items.Add(strname);
                        comboBoxCard.SelectedIndex = 0;
                        
                        CardData.IsWDG = true;
			            AIOWDM.WDGInit(CardNum);

                        btnDetectCOS.Enabled = false;

                        CardData.Base = Base;
                        found = true;
			            break;
                    case 0x02EE0: //PCIe-DIO-24s-CTR12
                        CardData.PPIs = 1;
                        
                        strname = String.Format("PCIe-DIO-24S-CTR12 Parallel Digital I/O Card");
                        comboBoxCard.Items.Add(strname);
                        comboBoxCard.SelectedIndex = 0;

                        CardData.Base = Base;
                        found = true;
                        break;
                  }// end switch
                  
                  // If card(s) are present but no valid cards were found:
                  if ((NumCards != 0) && (found == false))
                  {
                      //labelCardName.Text = "No Valid Card Found.";
                      strname = String.Format("No Valid Card Found");
                      comboBoxCard.Items.Add(strname);
                      comboBoxCard.SelectedIndex = 0;

                      textBoxMessage.Text = "No valid card was found!";
                      textBoxMessage.Text += "This may mean the card is not installed. ";
                      textBoxMessage.Text += "Check Device Manager for a card and its status";
                      textBoxMessage.Text += "You may consider rebooting your system.";
                      //comboBoxPort.Enabled = false;
                      //RunFlag = false;
                  }
		          
                  if (CardData.PPIs != 0)
		          {
			        CardData.Log =
			           ("AIOWDM provides generic IRQ-handling functions for all cards "
				        +"and special IRQ-handling functions for COS cards like this one. \r\n\r\n"
				        +"Click 'Detect Generic IRQs' to handle IRQs generically for the card selected above, "
				        +"or 'Detect COS IRQs' to handle IRQs for the card selected above by updating COS data. \r\n\r\n");
		          }       
		          else 
		          {
			          if (CardData.IsWDG)
			          {
				        CardData.Log =
				           ("AIOWDM provides generic IRQ-handling functions for all cards "
					        +"and special functions for watchdog cards like this one.\r\n\r\n"
					        +"Click 'Detect Generic IRQs' to handle IRQs generically for the card selected above, "
					        +"or 'Handle Watchdog IRQ' to handle the next IRQ for the card selected above with the selected action.\r\n\r\n"
					        +"You can also click 'Test Watchdog Timing' to set the watchdog timer, pet it a few times, then let it time out, "
					        +"'Read Temp' to read the temperature(if installed),"
					        +"'Read Status' to read the card''s status byte,"
					        +"or 'Reboot' to quickly reboot your computer.\r\n\r\n");
			          }
			          else 
			          {
				        CardData.Log =
					        ("AIOWDM provides generic IRQ-handling functions for all cards. \r\n\r\n"
					         +"Click 'Detect Generic IRQs' to handle IRQs for the card selected above.\r\n\r\n");
			          }
                  }
		        }// end for
	
	        }// end else


            //if (RunFlag)
            //{
                // This is the result setting the initial offset to be used:	  
                //Offset = PortOffsets[0];  //default to first valid base (Port 0)
                //CardData.IsSelected = true;
                //CardNum--;  // decriment after last increment so its the index of last valid card found
            //}

            // Clear/Enable COS Interrupt
            //uint tempAddr = (uint)CardData.Base + 11;
            //AIOWDM.OutPortB(tempAddr, 0x00);
            
            //Force to 0 we will only support a single card at this time
            CardNum = 0;
            //AIOWDM.RelOutPortB(CardNum, 10, 0x00); 
            //AIOWDM.RelOutPortB(CardNum, 11, 0x00);

            // Enable COS Interrupt
            //tempAddr = (uint)CardData.Base + 10;
            //AIOWDM.OutPortB(tempAddr, 0x00);
        }

        private void FindCardsWDM()
        {
 
            bool found = false;
            //char[] AddStr = new char[256];

            // Vars for QueryCardInfo() :
            UInt32 DeviceID, Base;
            UInt32 NameSize = 256;
            UInt16[] Name = new UInt16[256];
            String strname;

            // Init:
            RunFlag = true;

            // Set CardData to 0 as needed:
            //CardData.IsValid = false;
            //CardData.IsSelected = false;
            //CardData.DeviceID = 0;
            CardData.Base = 0;

            // Get the total number of cards installed:
            NumCards = AIOWDM.GetNumCards();

            if (NumCards == 0) // no cards present
            {
                strname = String.Format("No Card Found");
                comboBoxCard.Items.Add(strname);
                comboBoxCard.SelectedIndex = 0;

                textBoxMessage.Text = "No cards were found!";
                textBoxMessage.Text += "This may mean the card is not installed. ";
                textBoxMessage.Text += "Check Device Manager for a card and its status";
                textBoxMessage.Text += "You may consider rebooting your system.";
                comboBoxCard.Enabled = false;

                RunFlag = false;
            }
            else
            {
                // Loop through the cards and validate and store card data:
                for (CardNum = 0; CardNum != NumCards; CardNum++)
                {
                    // Validate and store card data:
                    Status = AIOWDM.QueryCardInfo(CardNum, out DeviceID, out Base, out NameSize, out strname);

                    // Populate list box with addresses found set flags:
                    switch (DeviceID)
                    {
                        case 0x0E50: //PCI-DIO-24S
                            CardData.PPIs = 1;
                            break;
                        case 0x0E56: //PCI-DIO-24DCS
                            CardData.PPIs = 1;
                            break;
                        case 0x0E60: //PCI-DIO-48S
                            CardData.PPIs = 2;

                            strname = String.Format("PCI-DIO-48S Parallel Digital I/O Card");
                            comboBoxCard.Items.Add(strname);
                            comboBoxCard.SelectedIndex = 0;

                            CardData.Base = Base;
                            found = true;
                            break;
                        case 0x22C0:
                        case 0x25C0:
                        case 0x2FC0:
                        case 0x2FC1: //PCI-WDG-CSM, P104-WDG-E, P104-WDG-CSM, and P104-WDG-CSMA
                            CardData.PPIs = 1;

                            strname = String.Format("PCI-WDG-CSM or P104 WDG Board");  // Port 0 ??
                            comboBoxCard.Items.Add(strname);
                            comboBoxCard.SelectedIndex = 0;

                            CardData.IsWDG = true;
                            AIOWDM.WDGInit(CardNum);

                            btnDetectCOS.Enabled = false;

                            CardData.Base = Base;
                            found = true;
                            break;
                        case 0x02EE0: //PCIe-DIO-24s-CTR12
                            CardData.PPIs = 1;

                            strname = String.Format("PCIe-DIO-24S-CTR12 Parallel Digital I/O Card");
                            comboBoxCard.Items.Add(strname);
                            comboBoxCard.SelectedIndex = 0;

                            CardData.Base = Base;
                            found = true;
                            break;
                    }
                }// end loop on validate present cards
            }//end else card present

            // If card(s) are present but no valid cards were found:
            if ((NumCards != 0) && (found == false))
            {
                strname = String.Format("No Valid Card Found");
                comboBoxCard.Items.Add(strname);
                comboBoxCard.SelectedIndex = 0;

                textBoxMessage.Text = "No valid card was found!";
                textBoxMessage.Text += "This may mean the card is not installed. ";
                textBoxMessage.Text += "Check Device Manager for a card and its status";
                textBoxMessage.Text += "You may consider rebooting your system."; 
                
                comboBoxCard.Enabled = false;

                RunFlag = false;
            }

            // once done set values if single card
            if (RunFlag)
            {
                // This is the result setting the initial offset to be used:	  
                //Offset = PortOffsets[0];  //default to first valid base (Port 0)
                Offset = 0;  //default to first valid base (Port 0)
                //CardData.IsSelected = true;
                //CardNum--;  // decriment after last increment so its the index of last valid card found
            }

            
            // Display card specific information:
            if (CardData.PPIs != 0)
            {
                CardData.Log =
                   ("AIOWDM provides generic IRQ-handling functions for all cards "
                    + "and special IRQ-handling functions for COS cards like this one. \r\n\r\n"
                    + "Click 'Detect Generic IRQs' to handle IRQs generically for the card selected above, "
                    + "or 'Detect COS IRQs' to handle IRQs for the card selected above by updating COS data. \r\n\r\n");
            }
            else
            {
                if (CardData.IsWDG)
                {
                    CardData.Log =
                       ("AIOWDM provides generic IRQ-handling functions for all cards "
                        + "and special functions for watchdog cards like this one.\r\n\r\n"
                        + "Click 'Detect Generic IRQs' to handle IRQs generically for the card selected above, "
                        + "or 'Handle Watchdog IRQ' to handle the next IRQ for the card selected above with the selected action.\r\n\r\n"
                        + "You can also click 'Test Watchdog Timing' to set the watchdog timer, pet it a few times, then let it time out, "
                        + "'Read Temp' to read the temperature(if installed),"
                        + "'Read Status' to read the card''s status byte,"
                        + "or 'Reboot' to quickly reboot your computer.\r\n\r\n");
                }
                else
                {
                    CardData.Log =
                        ("AIOWDM provides generic IRQ-handling functions for all cards. \r\n\r\n"
                         + "Click 'Detect Generic IRQs' to handle IRQs for the card selected above.\r\n\r\n");
                }
            }

        }
        
        //private static void MyThreadProc()
        private void MyThreadProc()
        {
            UInt32 retval = 0;

            while (true)
            {
                retval = AIOWDM.WaitForIRQ(CardNum);

                if (retval != 0)
                {
                    if (!bAbort) // If you exit Main could be gone before this thread is done
                        ReportIRQ();
                }
                else
                {
                    if (!bAbort) // If you exit Main could be gone before this thread is done
                    {
                        ReportError();
                    }
                    break;
                }
            }
            // Once you get here and leave this function the thread will exit:

        }

        // Updates the GUI with ERR codes:
        private void UpdateErrorText(string text)
        {
            // Set the GUI text.
            textBoxMessage.Text = text;
        }

        // Must use threadsafe delegate across threads:
        public delegate void UpdateErrTextCallback(string text);

        //private static void ReportError()
        private void ReportError()
        {   
            // Wait until it is safe to enter.
            mut.WaitOne();
            
            LastError = Marshal.GetLastWin32Error(); 

            // These are Windows Error codes used for Errors and expected status messages:
            int ERROR_SUCCESS = 0;
            int ERROR_INVALID_FUNCTION = 1;
            int ERROR_FILE_NOT_FOUND = 2;
            int ERROR_NO_MORE_FILES = 18;
            int ERROR_OPERATION_ABORTED = 995;
            
            int Error = LastError;
            string strMessage;

            if (Error == ERROR_OPERATION_ABORTED)
                strMessage = "IRQ request aborted.";
            else if (Error == ERROR_INVALID_FUNCTION)
                strMessage = "Error: this card already has a pending IRQ request. IRQ request aborted.";
            else
                strMessage = "Some error occurred in the driver. IRQ request aborted.";
 
            // Now from the thread you can safely call this:
            textBoxMessage.Invoke(new UpdateTextCallback(this.UpdateErrorText), new object[] { strMessage });
            
            // Release the Mutex.
            mut.ReleaseMutex(); 
        }

        // This method represents a resource that must be synchronized
        // so that only one thread at a time can enter.
        //private static void UseResource()
        private void ReportIRQ()
        {
            // Wait until it is safe to enter.
            mut.WaitOne();

            //Console.WriteLine("{0} has entered the protected area",
              //  Thread.CurrentThread.Name);

            // Access non-reentrant resources here.
            
            // Do some work.

            //CardData[CardNum].IRQCount++;
            numIRQ++;

            int count = numIRQ;
            ReceiveGenIRQ(count);
                       
            //Console.WriteLine("{0} is leaving the protected area\r\n",
              //  Thread.CurrentThread.Name);

            // Release the Mutex.
            mut.ReleaseMutex();
        }
/*
        public void ReceiveCOSData(int CardNum, TCOSData[] COSData)
        {
             //string formatString;
            //formatString = String.Format("{0,3:D}", numIRQ);
            //this.
            //labelGeneric.Text = formatString;           ulong I;
	        CComboBox* pCBox = (CComboBox*)GetDlgItem(IDC_COMBO_CARD); 

            for (I = 0; I < CardData[CardNum].PPIs; I++) {
               CardData[CardNum].LastCOSData[I].A = COSData[I].A;
               CardData[CardNum].LastCOSData[I].B = COSData[I].B;
               CardData[CardNum].LastCOSData[I].C = COSData[I].C;
	        }
            if (CardNum == pCBox->GetCurSel()) OnCbnSelchangeComboCard();
        }
*/

        // Updates the GUI:
        private void UpdateText(string text)
        {
            // Set the GUI text.
            labelGeneric.Text = text;            
        }

        // Must use threadsafe delegate across threads:
        public delegate void UpdateTextCallback(string text);
       
        public void ReceiveGenIRQ(int count)
        {
            // If you exit App, Main thread could be gone before this thread is done 
            // so cant go here on exit because this calls into the main thread
            
            string formatString;
            formatString = String.Format("{0,3:D}", count);
            
            // Now from the thread you can safely call this:
            labelGeneric.Invoke(new UpdateTextCallback(this.UpdateText),
                new object[] { formatString });
        }

        private void comboBoxAddress_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Clear/Enable COS Interrupt
            uint tempAddr = (uint)CardData.Base + 11;
            AIOWDM.OutPortB(tempAddr, 0x00);
        }

        private void btnDetectGeneric_Click(object sender, EventArgs e)
        {
            // Simple threading scenario:  Start a static method running
            // on a second thread.
            
            // Enable COS Interrupt
            //tempAddr = (uint)CardData.Base + 10;
            //AIOWDM.OutPortB(tempAddr, 0x00);
            
            // Clear/Enable COS Interrupt
            uint tempAddr = (uint)CardData.Base + 11;
            AIOWDM.OutPortB(tempAddr, 0x00);
            
            //AIOWDM.RelOutPortB(CardNum, 10, 0x00); 
            //AIOWDM.RelOutPortB(CardNum, 11, 0x00);
           
            // Create the thread(s) that will use the protected resource.
            bAbort = false;  // lower flag                    
            myThread = new Thread(new ThreadStart(MyThreadProc));
            myThread.Name = String.Format("Thread{0}", 1); 
            myThread.Start();

            textBoxMessage.Text = "Detecting generic IRQs; generate IRQs to see this sample count them. Press 'Abort IRQ Detect' to cancel when done.";
        }

        private void btnDetectCOS_Click(object sender, EventArgs e)
        {
            LastError = Marshal.GetLastWin32Error();
            ReportError(); 
            /*
            uint retval = AIOWDM.WaitForIRQ(CardNum - 1);
            if (retval != 0)
            {
                if (!bAbort) // If you exit Main could be gone before this thread is done
                {
                    ReportIRQ();
                }
            }
            else
            {
                if (!bAbort) // If you exit Main could be gone before this thread is done
                {
                    LastError = Marshal.GetLastWin32Error();
                    ReportError();
                }
                //break;
            }
            */
            //LastError = Marshal.GetLastWin32Error(); 
            textBoxMessage.Text = "Detecting COS IRQs; change the input data to see the COS data update. Press 'Abort IRQ Detect' to cancel when done.";
        }

        private void btnAbort_Click(object sender, EventArgs e)
        {
            UInt32 retval = 0;
            retval = AIOWDM.AbortRequest(0);  // abort wait state on the card and exit win thread 
            if (retval != 0)
            {
                textBoxMessage.Text = "Aborting IRQ detection...";          
            }
            else
            {
                textBoxMessage.Text = "No IRQ request to abort.";          
            }
        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            // Wait until it is safe to enter.
            mut.WaitOne();

            bAbort = true; 
            
            UInt32 retval = 0;
            retval = AIOWDM.AbortRequest(0);  // abort wait state and exit thread            

            // Release the Mutex.
            mut.ReleaseMutex();

            // Close the form and Application:
            this.Close();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // If user exits with system x button on main window and thread is running:

            // Wait until it is safe to enter.
            mut.WaitOne();

            bAbort = true;

            UInt32 retval = 0;
            retval = AIOWDM.AbortRequest(0);  // abort wait state and exit thread            

            // Release the Mutex.
            mut.ReleaseMutex();           
        }
    
    }// end class
}
